@@ -57,21 +57,17 @@ from theano.compile import \
     SymbolicOutput, Out, \
     Mode, \
     predefined_modes, predefined_linkers, predefined_optimizers, \
-    FunctionMaker, function, OpFromGraph, \
-    Component, External, Member, Method, \
-    Composite, ComponentList, ComponentDict, Module, \
+    FunctionMaker, function, function_dump, OpFromGraph, \
     ProfileMode, ProfileStats, \
-    Param, shared
+    Param, shared, as_op
 
 from theano.misc.safe_asarray import _asarray
 
-FancyModule = Module
-
 from theano.printing import pprint, pp
 
 from theano.scan_module import scan, map, reduce, foldl, foldr, clone
 
-from theano.updates import Updates, OrderedUpdates
+from theano.updates import OrderedUpdates
 
 # scan_module import above initializes tensor and scalar making these imports redundant
 #import tensor
@@ -79,7 +75,17 @@ from theano.updates import Updates, OrderedUpdates
 #we don't import by default as we don't want to force having scipy installed.
 #import sparse
 
-from theano.gradient import Rop, Lop, grad
+from theano.gradient import Rop, Lop, grad, subgraph_grad
+
+# This need to be before the init of GPU, as it add config variable
+# needed during that phase.
+import theano.tests
+if hasattr(theano.tests, "TheanoNoseTester"):
+    test = theano.tests.TheanoNoseTester().test
+else:
+    def test():
+        raise ImportError("The nose module is not installed."
+                          " It is needed for Theano tests.")
 
 if config.device.startswith('gpu') or config.init_gpu_device.startswith('gpu'):
     import theano.sandbox.cuda
@@ -186,20 +192,4 @@ def sparse_grad(var):
     return ret
 
 
-import theano.tests
-if hasattr(theano.tests, "TheanoNoseTester"):
-    test = theano.tests.TheanoNoseTester().test
-else:
-    def test():
-        raise ImportError("The nose module is not installed."
-                          " It is needed for Theano tests.")
-
-# This cannot be done in tensor/__init__.py due to a circular dependency -- randomstreams
-# depends on raw_random which depends on tensor.  As a work-around, we import RandomStreams
-# here and inject an instance in tensor.
-from theano import tensor
-from theano.tensor.randomstreams import RandomStreams
-# Imitate the numpy.random symbol with a tensor.random one
-tensor.random = RandomStreams(seed=0xBAD5EED, no_warn=True)
-del RandomStreams
 __import__('theano.tensor.shared_randomstreams')
