@@ -2,13 +2,14 @@ import atexit
 import copy
 import os
 import time
+import warnings
 
 import theano
 from theano.gof.link import WrapLinker
 from theano.compile.mode import (Mode, register_mode,
                                  predefined_modes, predefined_linkers,
                                  predefined_optimizers)
-from theano.gof.python25 import any
+from theano.compat.python2x import any
 from theano import gof
 from theano.configparser import config, AddConfigVar, IntParam, BoolParam
 from theano.compile.function_module import FunctionMaker
@@ -98,6 +99,10 @@ class Profile_Maker(FunctionMaker):
             # Lazy import to avoid compilation when importing theano.
             from theano.gof.cutils import run_cthunk
 
+        warnings.warn(
+            "DEPRECATION WARNING: The ProfileMode is deprecated. Use the Theano"
+            " flags/parameter to theano.function 'profile=True' instead"
+            " of 'mode=ProfileMode'")
         return ret
 
 
@@ -202,8 +207,6 @@ class ProfileMode(Mode):
 
         self.call_time = 0
         self.fn_time = 0
-        self.optimizer_time = 0
-        self.linker_time = 0
 
     def print_summary(self, **kwargs):
         """ Print 3 summary that show where the time is spend. The first show an Apply-wise summary, the second show an Op-wise summary, the third show an type-Op-wise summary.
@@ -289,9 +292,18 @@ class ProfileMode(Mode):
         apply_cimpl = self.apply_cimpl and other.apply_cimpl
         message = self.message
         variable_shape = diff_dict(self.variable_shape, other.variable_shape)
-        other_time = {'linker_time': self.linker_time - other.linker_time,
-                      'optimizer_time': self.optimizer_time -
-                                        other.optimizer_time}
+        self_linker_time = sum([ps.linker_time for ps
+                                 in self.profile_stats.values()])
+        other_linker_time = sum([ps.linker_time for ps
+                                 in other.profile_stats.values()])
+        self_optimizer_time = sum([ps.optimizer_time for ps
+                                 in self.profile_stats.values()])
+        other_optimizer_time = sum([ps.optimizer_time for ps
+                                 in other.profile_stats.values()])
+
+        other_time = {'linker_time': self_linker_time - other_linker_time,
+                      'optimizer_time': self_optimizer_time -
+                                        other_optimizer_time}
         self.print_summary_("print_diff_summary", compile_time,
                             fct_call_time, fct_call,
                             apply_time, apply_cimpl, message, variable_shape,
