@@ -4,19 +4,21 @@ fg.py: fg stands for FunctionGraph
 Contains the FunctionGraph class and exception
 types that it can raise
 """
+import StringIO
 import sys
 import time
+import traceback
 
 import theano
 from theano.gof import graph
 from theano.gof import utils
 from theano.gof import toolbox
-from theano.gof.python25 import all
+from theano.compat.python2x import all
 from theano import config
 import warnings
 NullType = None
 
-from theano.gof.python25 import OrderedDict
+from theano.compat.python2x import OrderedDict
 from theano.misc.ordered_set import OrderedSet
 
 
@@ -85,6 +87,11 @@ class FunctionGraph(utils.object2):
         #TODO: document what variables are[not] set in the FunctionGraph when a feature
         is added via the constructor.  How constructed is the FunctionGraph?
 
+        Note: the intermediate nodes between 'inputs' and 'outputs' are not explicitely
+        passed.
+
+        :param inputs: inputs nodes of the graph, usually declared by the user
+        :param outputs: outputs nodes of the graph.
         :param clone: If true, we will clone the graph. This is
         useful to remove the constant cache problem.
 
@@ -328,18 +335,29 @@ class FunctionGraph(utils.object2):
                             #if there is no path then r isn't really a graph input so we shouldn't be running error
                             #handler code in the first place
                             assert path is not None
+                            tr = getattr(r.tag, 'trace', None)
+                            detailed_err_msg = ""
+                            if tr:
+                                sio = StringIO.StringIO()
+                                traceback.print_list(tr, sio)
+                                tr = sio.getvalue()
+                                detailed_err_msg += "\nBacktrace when the variable is created:\n"
+                                detailed_err_msg += str(tr)
 
-                            raise MissingInputError((
+                            raise MissingInputError(
                                 'A variable that is an input to the graph was '
                                 'neither provided as an input to the function '
                                 'nor given a value. A chain of variables '
                                 'leading from this input to an output is %s. '
-                                'This chain may not be unique' % str(path)))
+                                'This chain may not be unique' % str(path) +
+                                detailed_err_msg)
 
                         #Standard error message
                         raise MissingInputError((
                             "An input of the graph, used to compute %s, "
-                            "was not provided and not given a value"
+                            "was not provided and not given a value."
+                            "Use the Theano flag exception_verbosity='high',"
+                            "for more information on this error."
                             % str(node)),
                             r)
 
@@ -444,12 +462,22 @@ class FunctionGraph(utils.object2):
         if verbose:
             print reason, r, new_r
         if r.fgraph is not self:
-            raise Exception("Cannot replace %s because it does not belong to this FunctionGraph" % r, str(reason))
-        if not r.type == new_r.type:
-            raise TypeError("The type of the replacement must be the same as the type of the original Variable.", r, new_r, r.type, new_r.type, str(reason))
+            raise Exception("Cannot replace %s because it does not belong "
+                            "to this FunctionGraph" % r, str(reason))
+        if r.type != new_r.type:
+            new_r2 = r.type.convert_variable(new_r)
+            # We still make sure that the type converts correctly
+            if new_r2 is None or new_r2.type != r.type:
+                raise TypeError("The type of the replacement must be "
+                                "compatible with the type of the original "
+                                "Variable.", r, new_r, r.type, new_r.type,
+                                str(reason))
+            new_r = new_r2
         if r not in self.variables:
-            # this variable isn't in the graph... don't raise an exception here, just return silently
-            # because it makes it easier to implement some optimizations for multiple-output ops
+            # this variable isn't in the graph... don't raise an
+            # exception here, just return silently because it makes it
+            # easier to implement some optimizations for
+            # multiple-output ops
             return
 
         if theano.config.compute_test_value != 'off':
@@ -483,11 +511,6 @@ class FunctionGraph(utils.object2):
         for r, new_r in pairs:
             self.replace(r, new_r, reason=reason)
 
-    def extend(self, feature):
-        warnings.warn("FunctionGraph.extend is deprecatd. It has been "
-                      "renamed to FunctionGraph.attach_feature")
-        return self.attach_feature(feature)
-
     def attach_feature(self, feature):
         """
         Adds a gof.toolbox.Feature to this function_graph
@@ -636,23 +659,6 @@ class FunctionGraph(utils.object2):
         """WRITEME Same as len(self.clients(r))."""
         return len(self.clients(r))
 
-    def nodes_getter(self):
-        warnings.warn("FunctionGraph.nodes is deprecated, it has been renamed 'apply_nodes'",
-                stacklevel=2)
-        return self.apply_nodes
-
-    def nodes_setter(self, value):
-        warnings.warn("FunctionGraph.nodes is deprecated, it has been renamed 'apply_nodes'",
-                stacklevel=2)
-        self.apply_nodes = value
-
-    def nodes_deleter(self):
-        warnings.warn("FunctionGraph.nodes is deprecated, it has been renamed 'apply_nodes'",
-                stacklevel=2)
-        del self.apply_nodes
-
-    nodes = property(nodes_getter, nodes_setter, nodes_deleter)
-
     def check_integrity(self):
         """WRITEME
         Call this for a diagnosis if things go awry.
@@ -711,17 +717,42 @@ class FunctionGraph(utils.object2):
         return self.__str__()
 
     ### clone ###
-    def clone(self):
+    def clone(self, check_integrity=True):
         """WRITEME"""
-        return self.clone_get_equiv()[0]
+        return self.clone_get_equiv(check_integrity)[0]
 
-    def clone_get_equiv(self):
+    def clone_get_equiv(self, check_integrity=True):
         """WRITEME"""
         equiv = graph.clone_get_equiv(self.inputs, self.outputs)
-        self.check_integrity()
+        if check_integrity:
+            self.check_integrity()
         e = FunctionGraph([equiv[i] for i in self.inputs],
                           [equiv[o] for o in self.outputs])
-        e.check_integrity()
+        if check_integrity:
+            e.check_integrity()
         for feature in self._features:
             e.attach_feature(feature)
         return e, equiv
+
+    def __getstate__(self):
+        """This is needed as some feature introduce instancemethod and
+        this is not picklable.
+        """
+        d = self.__dict__.copy()
+        for feature in self._features:
+            for attr in getattr(feature, "pickle_rm_attr", []):
+                del d[attr]
+        # The class Updater take fct as parameter and they are lambda function, so unpicklable.
+
+        # execute_callbacks_times have reference to optimizer, and they can't
+        # be pickled as the decorators with parameters aren't pickable.
+        if "execute_callbacks_times" in d:
+            del d["execute_callbacks_times"]
+
+        return d
+
+    def __setstate__(self, dct):
+        self.__dict__.update(dct)
+        for feature in self._features:
+            if hasattr(feature, "unpickle"):
+                feature.unpickle(self)
