@@ -2,9 +2,11 @@
 import unittest
 
 from nose.plugins.skip import SkipTest
+from nose.plugins.attrib import attr
 
+import theano
 from theano.gof.link import PerformLinker
-from theano.gof.cc import *
+from theano.gof.cc import CLinker, DualLinker, OpWiseCLinker
 from theano.gof.type import Type
 from theano.gof.graph import Variable, Apply, Constant
 from theano.gof.op import Op
@@ -20,7 +22,7 @@ class TDouble(Type):
     def filter(self, data):
         return float(data)
 
-    def c_declare(self, name, sub):
+    def c_declare(self, name, sub, check_input=True):
         return "double %(name)s; void* %(name)s_bad_thing;" % locals()
 
     def c_init(self, name, sub):
@@ -33,7 +35,7 @@ class TDouble(Type):
     def c_literal(self, data):
         return str(data)
 
-    def c_extract(self, name, sub):
+    def c_extract(self, name, sub, check_input=True):
         return """
         if (!PyFloat_Check(py_%(name)s)) {
             PyErr_SetString(PyExc_TypeError, "not a double!");
@@ -227,6 +229,17 @@ def test_clinker_dups():
     # note: for now the behavior of fn(2.0, 7.0) is undefined
 
 
+def test_clinker_not_used_inputs():
+    if not theano.config.cxx:
+        raise SkipTest("G++ not available, so we need to skip this test.")
+    # Testing that unused inputs are allowed.
+    x, y, z = inputs()
+    e = add(x, y)
+    lnk = CLinker().accept(Env([x, y, z], [e]))
+    fn = lnk.make_function()
+    assert fn(2.0, 1.5, 1.0) == 3.5
+
+
 def test_clinker_dups_inner():
     if not theano.config.cxx:
         raise SkipTest("G++ not available, so we need to skip this test.")
@@ -242,6 +255,7 @@ def test_clinker_dups_inner():
 # Test OpWiseCLinker #
 ######################
 
+# slow on linux, but near sole test and very central
 def test_opwiseclinker_straightforward():
     x, y, z = inputs()
     e = add(mul(add(x, y), div(x, y)), bad_sub(bad_sub(x, y), z))
@@ -253,6 +267,7 @@ def test_opwiseclinker_straightforward():
         # The python version of bad_sub always return -10.
         assert fn(2.0, 2.0, 2.0) == -6
 
+
 def test_opwiseclinker_constant():
     x, y, z = inputs()
     x = Constant(tdouble, 7.2, name='x')
