@@ -1,7 +1,9 @@
+import pickle
 import unittest
 
 import theano
 from theano.gof import CachedConstantError, FunctionGraph
+from theano import tensor as tt
 
 
 class TFunctionGraph(unittest.TestCase):
@@ -15,3 +17,10 @@ class TFunctionGraph(unittest.TestCase):
         v = theano.tensor.constant(1)
         assert v.cached
         FunctionGraph([], [v + 1])
+
+    def test_pickle(self):
+        v = tt.vector()
+        func = theano.gof.FunctionGraph([v], [v + 1])
+
+        s = pickle.dumps(func)
+        func2 = pickle.loads(s)
