@@ -1,6 +1,4 @@
-
-from kron import kron
 from ops import (cholesky, matrix_inverse, solve,
         diag, extract_diag, alloc_diag,
-        det, psd, eig, eigh,
+        det, psd, eig, eigh, eigvalsh,
         trace, spectral_radius_bound)
