@@ -13,7 +13,7 @@ import theano
 import theano.sandbox.rng_mrg
 from theano import tensor
 from theano.compile.pfunc import rebuild_collect_shared
-from theano.gof.python25 import any
+from theano.compat.python2x import any
 from theano.tests  import unittest_tools as utt
 from numpy.testing.noseclasses import KnownFailureTest
 
