@@ -45,6 +45,7 @@ __contact__ = "Razvan Pascanu <r.pascanu@gmail>"
 import itertools
 import logging
 import numpy
+import warnings
 
 from theano.compile import SharedVariable, function
 from theano import compile
@@ -54,7 +55,7 @@ from theano import tensor
 from theano import config
 from theano.updates import OrderedUpdates
 from theano.compile import ops
-from theano.gof.python25 import OrderedDict
+from theano.compat.python2x import OrderedDict
 
 
 from theano.scan_module import scan_op
@@ -74,7 +75,9 @@ def scan(fn,
          go_backwards=False,
          mode=None,
          name=None,
-         profile=False):
+         profile=False,
+         allow_gc=None,
+         strict=False):
     """
     This function constructs and applies a Scan op to the provided
     arguments.
@@ -309,6 +312,14 @@ def scan(fn,
         inner graph with the new cvm linker ( with default modes,
         other linkers this argument is useless)
 
+    :param allow_gc:
+        Set the value of allow gc for the internal graph of scan.  If
+        set to None, this will use the value of config.scan.allow_gc.
+
+    :param strict:
+        If true, all the shared variables used in ``fn`` must be provided as a
+        part of ``non_sequences`` or ``sequences``. 
+
     :rtype: tuple
     :return: tuple of the form (outputs, updates); ``outputs`` is either a
              Theano variable or a list of Theano variables representing the
@@ -382,7 +393,7 @@ def scan(fn,
     for i in xrange(n_seqs):
         if not isinstance(seqs[i], dict):
             seqs[i] = OrderedDict([('input', seqs[i]), ('taps', [0])])
-        elif seqs[i].get('taps', None):
+        elif seqs[i].get('taps', None) is not None:
             seqs[i]['taps'] = wrap_into_list(seqs[i]['taps'])
         elif seqs[i].get('taps', None) is None:
             # seqs dictionary does not have the ``taps`` key
@@ -393,7 +404,7 @@ def scan(fn,
         if outs_info[i] is not None:
             if isinstance(outs_info[i], dict):
                 # DEPRECATED :
-                if outs_info[i].get('return_steps', None):
+                if outs_info[i].get('return_steps', None) is not None:
                     raise ValueError(
                             "Using `return_steps` has been deprecated. "
                             "Simply select the entries you need using a "
@@ -404,14 +415,14 @@ def scan(fn,
             if not isinstance(outs_info[i], dict):
                 # by default any output has a tap value of -1
                 outs_info[i] = OrderedDict([('initial', outs_info[i]), ('taps', [-1])])
-            elif (not outs_info[i].get('initial', None) and
-                    outs_info[i].get('taps', None)):
+            elif (outs_info[i].get('initial', None) is None and
+                    outs_info[i].get('taps', None) is not None):
                 # ^ no initial state but taps provided
                 raise ValueError(('If you are using slices of an output '
                                   'you need to provide a initial state '
                                   'for it'), outs_info[i])
-            elif (outs_info[i].get('initial', None) and
-                  not outs_info[i].get('taps', None)):
+            elif (outs_info[i].get('initial', None) is not None and
+                  outs_info[i].get('taps', None) is None):
                 # ^ initial state but taps not provided
                 if 'taps' in outs_info[i]:
                     # ^ explicitly provided a None for taps
@@ -589,7 +600,9 @@ def scan(fn,
         if init_out.get('taps', None) == [-1]:
 
             actual_arg = init_out['initial']
-            arg = safe_new(init_out['initial'])
+            if not isinstance(actual_arg, tensor.Variable):
+                actual_arg = tensor.as_tensor_variable(actual_arg)
+            arg = safe_new(actual_arg)
             if isinstance(arg, tensor.Constant):
                 # safe new returns a clone of the constants, but that is not
                 # what we need for initial states
@@ -903,14 +916,29 @@ def scan(fn,
                              not isinstance(arg, tensor.Constant))]
 
     givens.update(OrderedDict(zip(other_scan_args, other_inner_args)))
-    other_shared_scan_args = [arg.variable for arg
-                        in dummy_f.maker.expanded_inputs
-                        if (isinstance(arg.variable, SharedVariable) and
-                            not arg.update)]
-    other_shared_inner_args = [safe_new(arg.variable, '_copy') for arg
-                        in dummy_f.maker.expanded_inputs
-                        if (isinstance(arg.variable, SharedVariable) and
-                            not arg.update)]
+
+    if strict:
+        non_seqs_set = set(non_sequences if non_sequences != None else [])
+
+        other_shared_scan_args = [arg.variable for arg
+                            in dummy_f.maker.expanded_inputs
+                            if (isinstance(arg.variable, SharedVariable) and
+                                not arg.update and
+                                arg.variable in non_seqs_set)]
+        other_shared_inner_args = [safe_new(arg.variable, '_copy') for arg
+                            in dummy_f.maker.expanded_inputs
+                            if (isinstance(arg.variable, SharedVariable) and
+                                not arg.update and
+                                arg.variable in non_seqs_set)]
+    else:
+        other_shared_scan_args = [arg.variable for arg
+                            in dummy_f.maker.expanded_inputs
+                            if (isinstance(arg.variable, SharedVariable) and
+                                not arg.update)]
+        other_shared_inner_args = [safe_new(arg.variable, '_copy') for arg
+                            in dummy_f.maker.expanded_inputs
+                            if (isinstance(arg.variable, SharedVariable) and
+                                not arg.update)]
     givens.update(OrderedDict(zip(other_shared_scan_args,
                            other_shared_inner_args)))
 
@@ -962,6 +990,8 @@ def scan(fn,
     ##
 
     tap_array = mit_sot_tap_array + [[-1] for x in xrange(n_sit_sot)]
+    if allow_gc is None:
+        allow_gc = config.scan.allow_gc
     info = OrderedDict()
 
     info['tap_array'] = tap_array
@@ -980,6 +1010,11 @@ def scan(fn,
     info['gpu'] = False
     info['as_while'] = as_while
     info['profile'] = profile
+    info['allow_gc'] = allow_gc
+    info['strict'] = strict
+    if strict:
+        warnings.warn('In the strict mode, all neccessary shared variables '
+                      'must be passed as a part of non_sequences', Warning)
 
     local_op = scan_op.Scan(inner_inputs, new_outs, info)
 
